// *-*-c++-*-* Emacs/c++-mode

#include<NTL/ZZX.h>
#include<NTL/new.h>

NTL_START_IMPL

/// The following functions are taken from NTL/ZZX.c
static inline
void KarSub(ZZ *T, const ZZ *b, long sb)
{
   long i;

   for (i = 0; i < sb; i++)
      sub(T[i], T[i], b[i]);
}

static inline
void KarAdd(ZZ *T, const ZZ *b, long sb)
{
   long i;

   for (i = 0; i < sb; i++)
      add(T[i], T[i], b[i]);
}

/// End of copy pasted functions from NTL/ZZX.c

/// Beginning of the transposed stuff

// for debugging only
void foo(const ZZ *v, long s) {
  for(long i=0; i < s; i++) cout << v[i] << " ";
}


/*
 * Transposition of plain multiplication
 *
 */

void TPlainMul(ZZX& b, long db, const ZZX& a, const ZZX& c)
{

  long da = deg(a);
  long dc = deg(c);

  if ((da >= 0 && dc > da + db) || (da < 0 && dc >= 0))
    Error("TPlainMul/ZZX: bad args");
 
  if (da < 0) {
    clear(b);
    return;
  }

  const ZZ *ap, *cp;
  ZZ *bp;
  
  ZZX la, lc;
  
  if (&b == &a) {
    la = a;
    ap = la.rep.elts();
  }
  else
    ap = a.rep.elts();
  
  if (&b == &c) {
    lc = c;
    cp = lc.rep.elts();
  }
  else
    cp = c.rep.elts();

  b.rep.SetLength(db+1);
  bp = b.rep.elts();
  
  long i, j, jmax;
  ZZ t, accum;
  
  for (i = 0; i <= db; i++) {
    jmax = min(dc, i+da);
    clear(accum);
    for (j = i; j <= jmax; j++) {
      mul(t, ap[j-i], cp[j]);
      add(accum, accum, t);
    }
    bp[i] = accum;
  }
  b.normalize();
}

static
void TPlainMul(ZZ *bp, long sb, const ZZ *ap, long sa, const ZZ *cp, long sc)
{
  long i, j, jmax;
  static ZZ t, accum;
  
  if ((sa > 0 && sc >= sa + sb) || (sa == 0 && sc != 0))
    Error("TPlainMul/ZZX: bad args");

  if (sa == 0) {
    for (i = 0; i < sb; i++) clear(bp[i]);
    return;
  }
  
  for (i = 0; i < sb; i++) {
    jmax = min(sc-1, i+sa-1);
    clear(accum);
    for (j = i; j <= jmax; j++) {
      mul(t, ap[j-i], cp[j]);
      add(accum, accum, t);
    }
    bp[i] = accum;
  }
}

/*
 *  Transposition of Karatsuba's multiplication
 *
 */

static inline
void KarFoldSub(ZZ *T, const ZZ *b, long sb, long hsa)
{
   long m = sb - hsa;
   long i;

   for (i = 0; i < m; i++)
      sub(T[i], b[hsa+i], b[i]);

   for (i = m; i < hsa; i++)
      negate(T[i], b[i]);
}

static inline
void KarAdd2(ZZ *T, const ZZ *a, const ZZ *b, long s)
{
   long i;
   for (i = 0; i < s; i++)
      add(T[i], a[i], b[i]);
}

static inline
void TKarMul(ZZ *b, long sb, const ZZ *a, long sa, const ZZ *c, long sc, ZZ *stk)
{

#ifdef DEBUG
  cout << "--> sa = " << sa << ", sb = " << sb << ", sc = " << sc << endl ;
  cout << "a = "; foo(a,sa); cout << endl;
  cout << "c = "; foo(c,sc); cout << endl;
#endif

  if ((sa > 0 && sc >= sa + sb) || (sa == 0 && sc != 0))
    Error("TKarMul/ZZX: bad args");

  if (sc <= 0) {
    for(long i=0 ; i < sb ; i++)
      clear(b[i]);
    return;
  }
  if (sb == 1) {
    if (sa == 1)
      mul(*b, *a, *c);
    else {
      clear(b[0]);
      for(long i=0 ; i < min(sa,sc) ; i++) {
	mul(stk[0], a[i], c[i]);
	add(b[0],b[0],stk[0]);
      }
    }
    return;
  }

  if (sa == 2 && sb == 2) {
    if (sc == 3) {
      add(b[0], c[0], c[1]);
      add(b[1], c[2], c[1]);
      mul(b[0], b[0], a[0]);
      mul(b[1], b[1], a[1]);
      sub(stk[0], a[0], a[1]);
      mul(stk[0], stk[0], c[1]);
      sub(b[0], b[0], stk[0]);
      add(b[1], b[1], stk[0]);
      return;
    }
    // now sc <= 2
    ZZ c0, c1;

    if (sc > 0) c0 = c[0];
    if (sc > 1) c1 = c[1];
    else clear(c1);

    add(b[0], c0, c1);
    b[1] = c1;
    mul(b[0], b[0], a[0]);
    mul(b[1], b[1], a[1]);
    sub(stk[0], a[0], a[1]);
    mul(stk[0], stk[0], c1);
    sub(b[0], b[0], stk[0]);
    add(b[1], b[1], stk[0]);
    return;
  }

  long hsa = (sa + 1) >> 1;
  long hsb = (sb + 1) >> 1;

#ifdef DEBUG
  cout << "hsa = " << hsa << endl;
  cout << "hsb = " << hsb << endl;
#endif
  if (sa >= sb && sb <= hsa) {
#ifdef DEBUG
    cout << "+ Degenerate case I" << endl;
#endif
    ZZ *T = stk; stk += sb;
    TKarMul(b, sb, a+hsa, sa-hsa, c+hsa, sc-hsa, stk);
    TKarMul(T, sb, a, hsa, c, min(sb+hsa-1,sc), stk);
    KarAdd(b, T, sb);
    return;
  }
  if (sa < sb && sa <= hsb) {
#ifdef DEBUG
    cout << "+ Degenerate case II" << endl;
#endif
    TKarMul(b, hsb, a, sa, c, min(sa+hsb-1,sc), stk);
    TKarMul(b+hsb, sb - hsb, a, sa, c+hsb, sc-hsb, stk);
    return;
  }

  long hs;
  if (sa < sb) hs = hsb; else hs = hsa;
  long hs2 = hs << 1;
  long hs3 = hs2 + hs;

#ifdef DEBUG
  cout << "+ Normal case" << endl;
#endif
  ZZ *T = stk; //stk += sc+1;
  stk += hs3;
  ZZ *T1 = T+hs;
  ZZ *T2 = T1+hs;

#ifdef DEBUG
    cout << "stk = " << (long)stk << endl;
#endif

  KarFoldSub(T, a, sa, hs);
  TKarMul(b, hs, T, hs, c+hs, min(hs2-1,sc-hs), stk);
  // cout << "b=";  foo(b,hs); cout << endl;

  long i = sc - hs;
  KarAdd2(T, c + hs, c, i);
  for(; i < min(sc, sa+sb-1-hs); ++i) T[i] = c[i];
  for(; i < sa+sb-1-hs; ++i) clear(T[i]);
  TKarMul(b+hs, sb - hs, a+hs, sa-hs, T1, sa+sb-1-hs2, stk);
  // cout << "b=";  foo(b,sb); cout << endl;

  KarSub(b+hs, b, sb-hs);
  // cout << "T="; foo(T,hs2-1); cout << endl;
  TKarMul(T2, hs, a, hs, T, hs2-1, stk);
  //  cout << "T2=";  foo(T2,hs); cout << endl;

  KarAdd(b, T2, hs);

#ifdef DEBUG
  cout << "b = "; foo(b,sb); cout << endl;
#endif
}

void TKarMul(ZZX& b, long db, const ZZX& a, const ZZX& c)
{

  vec_ZZ mem;
  
  const ZZ *ap, *cp;
  ZZ *bp;
  
  long sa = a.rep.length();
  long sc = c.rep.length();
  long sb = db + 1;
  
  if ((sa > 0 && sc >= sa + sb) || (sa == 0 && sc != 0))
    Error("TKarMul/ZZX: bad args");

  if (IsZero(a)) {
    clear(b);
    return;
  }
    
  if (&a == &b) {
    mem = a.rep;
    ap = mem.elts();
  }
  else
    ap = a.rep.elts();
  
  if (&c == &b) {
    mem = c.rep;
    cp = mem.elts();
  }
  else
    cp = c.rep.elts();
  
  b.rep.SetLength(sb);
  bp = b.rep.elts();
  
  long maxa, maxc, xover;
  
  maxa = MaxBits(a);
  maxc = MaxBits(c);
  xover = 2;
  
  if (sa < xover || sb < xover)
    TPlainMul(bp, sb, ap, sa, cp, sc);
  else {
    /* transposed karatsuba */
    // TODO << improve the size of stk?
    long n, hn, sp, depth;
    
    n = max(sa, sb);
    sp = 0;
    depth = 0;
    do {
      hn = (n+1) >> 1;
      sp += (hn << 2) - 1;
      n = hn;
      depth++;
    } while (n >= xover);
#ifdef DEBUG
    cout << "Allocating " << sp << " cells of size " 
	 << (((maxa + maxc + NumBits(sa) + 2*depth + 10)
	     + NTL_ZZ_NBITS-1)/NTL_ZZ_NBITS) << endl;
#endif
    ZZVec stk;
    stk.SetSize(sp,
		((maxa + maxc + NumBits(sa) + 2*depth + 10)
		 + NTL_ZZ_NBITS-1)/NTL_ZZ_NBITS);
    TKarMul(bp, sb, ap, sa, cp, sc, stk.elts());
  }
  b.normalize();
}

NTL_END_IMPL
