// *-*-c++-*-* Emacs/c++-mode

#include <NTL/ZZ_pX.h>
#include <NTL/ZZX.h>
#include <NTL/new.h>
#include "ZZX.h"

// copied from NTL
// These crossovers are tuned for a Pentium, but hopefully
// they should be OK on other machines as well.

const long SS_kbound = 40;
const double SS_rbound = 1.25;
// end of copy


NTL_START_IMPL

void TPlainMul(ZZ_pX& x, long dx, const ZZ_pX& a, const ZZ_pX& b)
{
  long da = deg(a);
  long db = deg(b);

  if ((da >= 0 && db > da + dx) || (da < 0 && db >= 0))
    Error("ZZ_pX/TPlainMul: bad args");

  if (da < 0) {
    clear(x);
    return;
  }
  
  if (da == 0) {
    mul(x, b, a.rep[0]);
    return;
  }

  const ZZ_p *ap, *bp;
  ZZ_p *xp;
  
  ZZ_pX la, lb;
  
  if (&x == &a) {
    la = a;
    ap = la.rep.elts();
  }
  else
    ap = a.rep.elts();
  
  if (&x == &b) {
    lb = b;
    bp = lb.rep.elts();
  }
  else
    bp = b.rep.elts();
  
  x.rep.SetLength(dx+1);
  
  xp = x.rep.elts();
  
  long i, j, jmax;
  static ZZ t, accum;
  
  for (i = 0; i <= dx; i++) {
    jmax = min(db,i+da);
    clear(accum);
    for (j = i; j <= jmax; j++) {
      mul(t, rep(ap[j-i]), rep(bp[j]));
      add(accum, accum, t);
    }
    conv(xp[i], accum);
  }
  x.normalize();
}

void TFFTMul(ZZ_pX& b, long db, const ZZ_pX& a, const ZZ_pX& c)
{
  long k, d;
  long da = deg(a);
  long dc = deg(c);

  if ((da >= 0 && dc > da + db) || (da < 0 && dc >= 0))
    Error("ZZ_pX/TFFTMul: bad args");

  if (da < 0) {
    clear(b);
    return;
  }
  
  if (da == 0) {
    mul(b, c, a.rep[0]);
    return;
  }

  // TODO << remove this reverse
  ZZ_pX la;
  reverse(la,a);
  
  d = da+db;
  k = NextPowerOfTwo(d+1);
  
  FFTRep R1(INIT_SIZE, k), R2(INIT_SIZE, k);

  ToFFTRep(R1, la, k);
  ToFFTRep(R2, c, k);
  mul(R1, R1, R2);
  FromFFTRep(b, R1, deg(a), d);
}

void tmul(ZZ_pX& b, long db, const ZZ_pX& a, const ZZ_pX& c)
{
  long da = deg(a);
  long dc = deg(c);
  if ((da >= 0 && dc > da + db) || (da < 0 && dc >= 0))
    Error("ZZ_pX/tmul: bad args");
 
  if (da < 0) {
    clear(b);
    return;
  }

  long k = ZZ_p::ModulusSize();
  long s = min(deg(a), db) + 1;
  
  if (s == 1 || (k == 1 && s < 40) || (k == 2 && s < 20) ||
      (k == 3 && s < 12) || (k <= 5 && s < 8) ||
      (k <= 12 && s < 4) )  {
    TPlainMul(b, db, a, c);
  }
  else if (s < 80) {
    ZZX A, B, C;
    conv(A, a);
    conv(C, c);
    TKarMul(B, db, A, C);
    conv(b, B);
  }
  else {
    long mbits;
    mbits = NumBits(ZZ_p::modulus());
    if (k >= SS_kbound && 
	SSRatio(deg(a), mbits, deg(b), mbits) < SS_rbound) {
      /*
      ZZX A, B, C;
      conv(A, a);
      conv(B, b);
      TSSMul(C, A, B);
      conv(c, C);
      */
      Error("ZZ_pX/tmul: not implemented yet");
    }
    else {
      TFFTMul(b, db, a, c);
    }
  }
}

void TPlainDivRem(ZZ_pX& b, long db, const ZZ_pX& a, const ZZ_pX& q, const ZZ_pX& r)
{
   long da, dq, dr, i, j, LCIsOne;
   ZZ_p const *qp, *rp, *ap;
   ZZ_p *bp;

   ZZ_p LCInv;

   da = deg(a);
   dq = deg(q);
   dr = deg(r);

#ifdef DEBUG
   cout << "da = " << da << ", db = " << db << ", dq = " << dq << ", dr = " << dr  << endl;
#endif
   if (da < 0) Error("ZZ_pX/TPlainDivRem: division by zero");
   if (db < dq + da) Error("ZZ_pX/TPlainDivRem: bad args");
   if (dr >= da) Error("ZZ_pX/TPlainDivRem: invalid arguments");

   ZZ_pX la;
   if (&b == &a) {
     la = a;
     ap = la.rep.elts();
   }
   else
     ap = a.rep.elts();

   ZZ_pX lq;
   if (&b == &q) {
     lq = q;
     qp = lq.rep.elts();
   }
   else
     qp = q.rep.elts();

   ZZ_pX lr;
   if (&b == &r) {
     lr = r;
     rp = lr.rep.elts();
   }
   else
     rp = r.rep.elts();

   if (IsOne(ap[da]))
      LCIsOne = 1;
   else {
      LCIsOne = 0;
      inv(LCInv, ap[da]);
   }

   // TODO << do better by resizing r instead copying
   b.rep.SetLength(db+1);
   bp = b.rep.elts();

   for (i = 0; i <= dr; i++)
      bp[i] = rp[i];

   for (i = dr + 1 ; i < da; i++)
      clear(bp[i]);

   ZZ s, t;
   for (i = da; i <= db; i++) {
     clear(t);
     for (j = da-1; j >= 0; j--) {
       mul(s, rep(ap[j]), rep(bp[i-da+j]));
       add(t, t, s);
     }
     sub(t, rep(qp[i-da]),t);
     conv(bp[i], t);

     if (!LCIsOne)
       mul(bp[i], bp[i], LCInv);
   }

   b.normalize();
}

void TPlainDiv(ZZ_pX& b, long db, const ZZ_pX& a, const ZZ_pX& q)
{
   long da, dq, dr, i, j, LCIsOne;
   ZZ_p const *qp, *rp, *ap;
   ZZ_p *bp;

   ZZ_p LCInv;

   da = deg(a);
   dq = deg(q);
   dr = da-1;
   db = dq + da;
   
#ifdef DEBUG
   cout << "da = " << da << ", db = " << db << endl;
#endif
   if (da < 0) Error("ZZ_pX/TPlainDiv: division by zero");
   if (db < dq + da) Error("ZZ_pX/TPlainDiv: bad args");

   ZZ_pX la;
   if (&b == &a) {
     la = a;
     ap = la.rep.elts();
   }
   else
     ap = a.rep.elts();

   ZZ_pX lq;
   if (&b == &q) {
     lq = q;
     qp = lq.rep.elts();
   }
   else
     qp = q.rep.elts();

   if (IsOne(ap[da]))
      LCIsOne = 1;
   else {
      LCIsOne = 0;
      inv(LCInv, ap[da]);
   }

   b.rep.SetLength(db+1);
   bp = b.rep.elts();

   for (i = 0; i < da; i++)
      clear(bp[i]);

   ZZ s, t;
   long da2 = da << 1;
   for (i = da; i <= db; i++) {
     clear(t);
     for (j = da-1; j >= da2-i; j--) {
       mul(s, rep(ap[j]), rep(bp[i-da+j]));
       add(t, t, s);
     }
     sub(t, rep(qp[i-da]),t);
     conv(bp[i], t);

     if (!LCIsOne)
       mul(bp[i], bp[i], LCInv);
   }

   b.normalize();
}

void TPlainRem(ZZ_pX& b, long db, const ZZ_pX& a, const ZZ_pX& r)
{
   long da, dq, dr, i, j, LCIsOne;
   ZZ_p const *rp, *ap;
   ZZ_p *bp;

   ZZ_p LCInv;

   da = deg(a);
   dr = deg(r);
   dq = db - da;
   
#ifdef DEBUG
   cout << "da = " << da << ", db = " << db << ", dr = " << dr << endl;
#endif
   if (da < 0) Error("ZZ_pX/TPlainRem: division by zero");
   if ((dr >= da) || (dq < 0)) Error("ZZ_pX/TPlainRem: invalid arguments");

   ZZ_pX la;
   if (&b == &a) {
     la = a;
     ap = la.rep.elts();
   }
   else
     ap = a.rep.elts();

   ZZ_pX lr;
   if (&b == &r) {
     lr = r;
     rp = lr.rep.elts();
   }
   else
     rp = r.rep.elts();

   if (IsOne(ap[da]))
      LCIsOne = 1;
   else {
      LCIsOne = 0;
      inv(LCInv, ap[da]);
   }

   // TODO << do better by resizing r instead copying
   b.rep.SetLength(db+1);
   bp = b.rep.elts();

   for (i = 0; i <= dr; i++)
      bp[i] = rp[i];

   for (i = dr + 1 ; i < da; i++)
      clear(bp[i]);

   ZZ s, t;
   for (i = da; i <= db; i++) {
     clear(t);
     for (j = da-1; j >= 0; j--) {
       mul(s, rep(ap[j]), rep(bp[i-da+j]));
       add(t, t, s);
     }
     negate(t, t);
     conv(bp[i], t);

     if (!LCIsOne)
       mul(bp[i], bp[i], LCInv);
   }

   b.normalize();
}

// TODO << ntl's includes?
void CopyReverse(ZZ_pX& x, const ZZ_pX& a, long lo, long hi);

inline
void TCyclicReduce(ZZ_pX& a, long n, long m, const ZZ_pX& x)

// computes the transpose of x = a mod X^m-1
// n = deg(a)

{
   long i, j, dx;
   ZZ_p accum;

   if (n < m) {
      a = x;
      return;
   }

   dx = deg(x);
   if (dx >= m) Error("TCyclicReduce: illegal arguments");

   ZZ_pX lx;
   const ZZ_p *xp;
   if (&x == &a) {
     lx = x;
     xp = lx.rep.elts();
   }
   else
     xp = x.rep.elts();

   a.rep.SetLength(n+1);

   for (i = 0; i <= dx; i++) a.rep[i] = xp[i];
   for (i = dx + 1; i < m; i++) clear(a.rep[i]);
   for (i = m; i <= n; i++) a.rep[i] = a.rep[i-m]; 

   a.normalize();
}

void TFFTDivRem(ZZ_pX& b, long db, const ZZ_pX& a, const ZZ_pX& q, const ZZ_pX& r)
{
   long dq = deg(q);
   long dr = deg(r);
   long da = deg(a);

   if (da < 0) Error("ZZ_pX/TFFTDivRem: division by zero");
   if ((dr >= da) || (db < dq + da)) Error("ZZ_pX/TFFTDivRem: illegal arguments");

   // TODO
   if (db >= 3*da) {
     Error("ZZ_pX/TFFTDivRem: not implemented yet");
   }

   long k = NextPowerOfTwo(2*(db-da)+1);
   long k1 = NextPowerOfTwo(da);
   long mx = max(k1, k);
   long l = 1L << k1;
#ifdef DEBUG
   cout << "l = " << l << endl;
#endif

   ZZ_pX P1, P2, P3;

   CopyReverse(P3, a, 0, da);
   InvTrunc(P2, P3, db - da + 1);

   FFTRep R1(INIT_SIZE, mx), R2(INIT_SIZE, mx);

   ToFFTRep(R1, r, k1, 0, dr);
   LeftShift(P3, P3, l-da);
   ToFFTRep(R2, P3, k1);
   mul(R1, R1, R2);
   FromFFTRep(P3, R1, 0, min(l - 1, db - da));
#ifdef DEBUG
   cout << "P3 = " << P3 << endl;
#endif
   TCyclicReduce(P1, db-da, l, P3);
   sub(P1, q, P1);
#ifdef DEBUG
   cout << "P1 = " << P1 << endl;
#endif
   ToFFTRep(R1, P1, k);
   ToFFTRep(R2, P2, k);
   mul(R1, R1, R2);
   FromFFTRep(P3, R1, 0, db - da);
   LeftShift(P3, P3, da);

#ifdef DEBUG
   cout << "P3 = " << P3 << endl;
#endif

   TCyclicReduce(b, db, l, r);
#ifdef DEBUG
   cout << "b = " << b << endl;
#endif
   add(b, b, P3);
}

void TFFTRem(ZZ_pX& b, long db, const ZZ_pX& a, const ZZ_pX& r)
{
   long m = deg(a);
   long n = db;
   long dq = db - m;
   long dr = deg(r);

   long k;

   if (m < 0) Error("ZZ_pX/TTFTRem: division by zero");
   if (dr >= m) Error("ZZ_pX/TFFTRem: illegal arguments");

   //TODO
   if (n >= 3*m) {
     Error("ZZ_pX/TFFTRem: not implemented yet");
   }

   ZZ_pX P1, P2, P3;

   CopyReverse(P3, a, 0, m);
   InvTrunc(P2, P3, n - m + 1);

   k = NextPowerOfTwo(2*(n-m)+1);
   long k1 = NextPowerOfTwo(m);
   long mx = max(k1, k);
   long l = 1L << k1;

#ifdef DEBUG
   cout << "l = " << l << endl;
#endif

   FFTRep R1(INIT_SIZE, mx), R2(INIT_SIZE, mx);

   ToFFTRep(R1, r, k1, 0, dr);
   LeftShift(P3, P3, l-m);
   ToFFTRep(R2, P3, k1);
   mul(R1, R1, R2);
   FromFFTRep(P3, R1, 0, min(l - 1, n - m));
#ifdef DEBUG
   cout << "P3 = " << P3 << endl;
#endif
   TCyclicReduce(P1, n-m, l, P3);
#ifdef DEBUG
   cout << "P1 = " << P1 << endl;
#endif
   ToFFTRep(R1, P1, k);
   ToFFTRep(R2, P2, k);
   mul(R1, R1, R2);
   FromFFTRep(P3, R1, 0, n - m);
   LeftShift(P3, P3, m);

#ifdef DEBUG
   cout << "P3 = " << P3 << endl;
#endif

   TCyclicReduce(b, n, l, r);
#ifdef DEBUG
   cout << "b = " << b << endl;
#endif
   sub(b, b, P3);
}

void TFFTDiv(ZZ_pX& b, long db, const ZZ_pX& a, const ZZ_pX& q)
{
   long m = deg(a);
   long n = db;
   long dq = deg(q);

   long k;

   if (m < 0) Error("ZZ_pX/TFFTDiv: division by zero");
   if (db < dq + m) Error("ZZ_pX/TFFTDiv: bad args");

   // TODO
   if (n >= 3*m) {
     Error("ZZ_pX/TFFTDiv: not implemented yet");
   }

   ZZ_pX P2, P3;

   CopyReverse(P3, a, 0, m);
   InvTrunc(P2, P3, n-m+1);

   k = NextPowerOfTwo(2*(n-m)+1);

   FFTRep R1(INIT_SIZE, k), R2(INIT_SIZE, k);

   ToFFTRep(R1, P2, k, 0, n-m);
   ToFFTRep(R2, q, k);
   mul(R1, R1, R2);
   FromFFTRep(b, R1, 0, n-m);
}

void TDivRem(ZZ_pX& b, long db, const ZZ_pX& a, const ZZ_pX& q, const ZZ_pX& r)
{
  if (deg(a) > NTL_ZZ_pX_DIV_CROSSOVER && deg(q) > NTL_ZZ_pX_DIV_CROSSOVER)
    TFFTDivRem(b, db, a, q, r);
  else
    TPlainDivRem(b, db, a, q, r);
}

void tdiv(ZZ_pX& b, long db, const ZZ_pX& a, const ZZ_pX& q)
{
   if (deg(a) > NTL_ZZ_pX_DIV_CROSSOVER && deg(q) > NTL_ZZ_pX_DIV_CROSSOVER)
      TFFTDiv(b, db, a, q);
   else
      TPlainDiv(b, db, a, q);
}

void trem(ZZ_pX& b, long db, const ZZ_pX& a, const ZZ_pX& r)
{
   if (deg(a) > NTL_ZZ_pX_DIV_CROSSOVER && (db - deg(a)) > NTL_ZZ_pX_DIV_CROSSOVER)
      TFFTRem(b, db, a, r);
   else
      TPlainRem(b, db, a, r);
}

void TMulMod(ZZ_pX& b, const ZZ_pX& a, const ZZ_pX& x, const ZZ_pX& f)
{
  // Transposed of x= ab mod f
  long m = deg(f);

  if (deg(a) >= m || deg(x) >= m || m < 0) 
    Error("ZZ_pX/TMulMod: bad args");
  
  ZZ_pX t;

  trem(t, 2*(m-1), f, x);  
  tmul(b, m-1, a, t);
}


NTL_END_IMPL

