// *-*-c++-*-* Emacs/c++-mode
/*
 * New faster FFTdiv.
 * NewtonInvTruncMul is used by (T)Fast2Eval.
 */

#include <NTL/ZZ_pX.h>
#include "ZZ_pX.h"

#ifndef NTL_OLD_ZZ_pX_MUL

#include <NTL/ZZX.h>

#endif

#include <NTL/new.h>

#undef NTL_ZZ_pX_NEWTON_CROSSOVER
#define NTL_ZZ_pX_NEWTON_CROSSOVER 2

NTL_START_IMPL
#ifdef TODO
void PlainDivTrunc(ZZ_pX& x, const ZZ_pX& b, const ZZ_pX& a, long m)
  
  /* x = (b/a) % X^m, input not output, constant term a is nonzero */
  
{
  long i, k, na, nb;
  static ZZ v, t;
  ZZ_p s;
  const ZZ_p* ap;
  const ZZ_p* bp;
  ZZ_p* xp;
  
  na = deg(a);
  nb = deg(b);
  
  if (na < 0) Error("division by zero");
  
  inv(s, ConstTerm(a));
  long is_one = IsOne(s);

  ap = a.rep.elts();
  bp = b.rep.elts();
  x.rep.SetLength(m);
  xp = x.rep.elts();

  if (na == 0) {  
    if is_one
      for (k = 0; k < min(m, nb+1); k++) xp[k] = bp[k];
    else
      for (k = 0; k < Min(m, nb+1); k++) mul(xp[k], bp[k], s);
    x.normalize();
    return;
  }
  
  for (k = 0; k < min(m, nb+1); k++) xp[k] = bp[k];
  for (; k < m; k++) clear(xp[k]);

  conv(t, s);  
  for (k = 1; k < m; k++) {
    clear(v);
    conv(v, xp[k]);
    if (!is_one)
      mul(v, v, t);
    xp[k] = v;
    for (i = k+1; i < min(m, na+1+k); i++) {
      mul(v, v, rep(ap[i-k]));
      sub(xp[i], xp[i], v);
    }
  } 
  x.normalize();
}
#endif

// TODO << check memory clash
void NewtonInvTruncMul(ZZ_pX& x, const ZZ_pX& a, long m, const ZZ_pX b)
{  /* b/a mod X^m */

#ifdef DEBUG
    cout << "--> NewtonInvTruncMul, m = " << m << endl;
    cout << "a = " << a << endl;
    cout << "b = " << b << endl;
#endif
  x.SetMaxLength(m);
  long i, t, k;
  
  long log2_newton = NextPowerOfTwo(NTL_ZZ_pX_NEWTON_CROSSOVER)-1;
  PlainInvTrunc(x, a, 1L << log2_newton);
  
  t = NextPowerOfTwo(m);
  
  FFTRep R1(INIT_SIZE, t), R2(INIT_SIZE, t);
  ZZ_pX P1(INIT_SIZE, m/2);
  
  long a_len = min(m, a.rep.length());
  
  ZZ_pXModRep a_rep;
  ToZZ_pXModRep(a_rep, a, 0, a_len-1);
  
  k = 1L << log2_newton; 
  t = log2_newton;
  long l;
  while (k < m) {
    l = min(2*k, m);
    if (l >= m) break;
    ToFFTRep(R1, x, t+1);
    ToFFTRep(R2, a_rep, t+1, 0, l-1); 
    mul(R2, R2, R1);
    FromFFTRep(P1, R2, k, l-1);
    
    ToFFTRep(R2, P1, t+1);
    mul(R2, R2, R1);
    FromFFTRep(P1, R2, 0, l-k-1);
    
    x.rep.SetLength(l);
    long y_len = P1.rep.length();
    for (i = k; i < l; i++) {
      if (i-k >= y_len)
	clear(x.rep[i]);
      else
	negate(x.rep[i], P1.rep[i-k]);
    }
    x.normalize();
    
    t++;
    k = l;
  }
#ifdef DEBUG
    cout << "x = " << x << endl;
    cout << "ax = " << (a*x) << endl;
    cout << "Last step, k =" << k << ", l =" << l << ", t=" << t << endl;
#endif
  FFTRep R3(INIT_SIZE, t);
  ZZ_pX P2l(INIT_SIZE, k);
  ZZ_pX P2h(INIT_SIZE, l-k);

  ToFFTRep(R1, x, t+1);

  ToFFTRep(R2, b, t+1, k, l-1);  
  mul(R3, R1, R2);
  FromFFTRep(P2h, R3, 0, l-k-1); // hight part

  ToFFTRep(R2, b, t+1, 0, k-1);  
  mul(R3, R1, R2);
  FromFFTRep(P2l, R3, 0, l-1); // low part

#ifdef DEBUG
    cout << "P2l= " << P2l << endl;
    cout << "P2h= " << P2h << endl;
    cout << "x*b = " << x*b << endl;
#endif

  ToFFTRep(R2, a_rep, t+1, 0, l-1); 
  mul(R2, R1, R2);
  FromFFTRep(P1, R2, k, l-1);

#ifdef DEBUG
    cout << "P1 = " << P1 << endl;
#endif

  /*   P1 = P1*P2l */
  ToFFTRep(R2, P1, t+1, 0, l-k-1); 
  ToFFTRep(R3, P2l, t+1, 0, l-k-1); 
  mul(R1, R2, R3);
  FromFFTRep(P1, R1, 0, l-k-1); // hight part

#ifdef DEBUG
    cout << "P1*P2l = " << P1 << endl;
#endif
 
  x.rep.SetLength(l);
  for (i = 0; i < k; i++) {
    x.rep[i]=P2l.rep[i];
  }
  for (i = k; i < l; i++) {
    add(x.rep[i], P2l.rep[i], P2h.rep[i-k]);
  }
  for (i = k; i < l; i++) {
    sub(x.rep[i], x.rep[i], P1.rep[i-k]);
  }
  x.normalize();
#ifdef DEBUG
    cout << "x = " << x << endl;
#endif
}

namespace dvlp {
void FFTDiv(ZZ_pX& q, const ZZ_pX& a, const ZZ_pX& b)
{
   long n = deg(b);
   long m = deg(a);
   long k;

   if (m < n) {
      clear(q);
      return;
   }

   if (m >= 3*n) {
      ZZ_pXModulus B;
      build(B, b);
      div(q, a, B);
      return;
   }

   ZZ_pX P1, P2, P3;

   CopyReverse(P1, a, 0, m);
   CopyReverse(P2, b, 0, n);
   NewtonInvTruncMul(P3, P2, m-n+1,P1);
   CopyReverse(q, P3, 0, m-n);
   q.normalize();
}

} // end ns dvlp
NTL_END_IMPL



