// *-*-c++-*-* Emacs/c++-mode

#include <assert.h>
#include <NTL/ZZ_pX.h>
#include "ZZ_pX.h"

NTL_CLIENT

static long debug;

static
void usage(int argc, char **argv) {
  cerr << argv[0] << " 0|1 (debug) nb of bits of p case" << endl;
  cerr << "case = 0 (mul) | 1 (div) | 2 (eval) | 3 (interp)" << endl;
}

static
void test_mul() {
  long N=5000000;
  ZZ_pX a, b, c, d, c1, b1, b2, b3;
  long n,m,i;
  double t0, t1, t2;

  cout << "# deg\tmul\ttmul\ttmul/mul" << endl;

  for(m = 1; m < 200; m++) {
    m += 3;
    random(a, m+1);
    random(b, m+1);

    SetCoeff(a,0,1);
    SetCoeff(a,m,1);
    SetCoeff(b,m,1);
    SetCoeff(b,0,1);

    mul(c1,a,b);

    TPlainMul(b1,m,a,c1);
    tmul(b2,m,a,c1);
    TFFTMul(b3,m,a,c1);

    assert((b1==b2));
    assert((b1==b3));

    if (debug == 1) continue;
    cout.flush();

    t0 = GetTime();
    for(i = 0 ; i < N / (m*m); i++) mul(c,a,b);
    t1 = (GetTime()-t0)*(m*m);

    t0 = GetTime();
    for(i = 0 ; i < N / (m*m); i++) tmul(b,m,a,c);
    t2 = (GetTime()-t0)*(m*m);

    cout <<  m << "\t" << t1 << "\t" << t2 << "\t" << t2/t1 << endl;
  }
}

static
void test_rem() {
  const long N=5000000;
  ZZ_pX a, b, c, b1, b2, b3;
  long n,m,i;
  double t0, t1, t2;

  cout << "# deg\trem\ttrem\ttrem/rem" << endl;
  
  for(m = 1; m < 200; m++) {
    if (m > 100) m+= 4;
    else m+=1;
    random(a, m+1);
    random(b, 2*m+1);

    SetCoeff(a,0,1);
    SetCoeff(a,m,1);
    SetCoeff(b,2*m,1);
    SetCoeff(b,0,1);

    rem(c,b,a);

    TPlainRem(b1,2*m,a,c);
    trem(b2,2*m,a,c);
    TFFTRem(b3,2*m,a,c);

    assert((b1==b2));
    assert((b1==b3));

    if (debug == 1) continue;
    cout.flush();

    t0 = GetTime();
    for(i = 0 ; i < N / (m*m); i++) rem(c,b,a);
    t1 = GetTime()-t0;

    t0 = GetTime();
    for(i = 0 ; i < N / (m*m); i++) trem(b,2*m,a,c);
    t2 = GetTime()-t0;

    cout <<  m << "\t" << t1 << "\t" << t2 << "\t" << t2/t1 << endl;
  }
}

void test_eval(void) {

  double t,t1,t2,t3,t4,t5,t6;

  ZZ_pX f, f1, f2, f3;
  vec_ZZ_p x, y1, y2, y3;
  long i, k, n, nit;

#ifdef DEBUG
  long N = 0;
#else
  long N = 500000;
#endif

  cout << "# deg\tPlainEval\tTPlainEval\tTPlainEval/PlainEval\tfast1eval\ttfast1eval\ttfast1eval/fast1eval\tfast2eval\ttfast2eval\ttfast2eval/fast2eval" << endl;

  for(n = 2; n < 300; n++) {
    if (n > 10) {
      n+=1;
      N = 50000;
    }
    if (n > 200) {
      n+=1;
      N = 5000;
    }

    x.SetLength(n);
    for(i = 0; i < n; i++) random(x[i]);

    random(f,n);
    SetCoeff(f,n-1,1);

#ifdef DEBUG
    cout << "x = " << x << endl;
    cout << "f = " << f << endl;
#endif

    if (debug) nit = 1;
    else nit = 1 + N/n;

    t = GetTime();
    for (i = 0; i < nit; i++) PlainEval(y1, f, x);
    t1 = (1000*(GetTime()-t)/nit);
#ifdef DEBUG
    cout << "y1 = " << y1 << endl;
#endif

    t = GetTime();
    for (i = 0; i < nit; i++) TPlainEval(f1, n-1, x, y1);
    t2 = (1000*(GetTime()-t)/nit);
#ifdef DEBUG
    cout << "f1 = " << f1 << endl;
#endif

    t = GetTime();
    for (i = 0; i < nit; i++) Fast1Eval(y2, f, x);
    t3 = (1000*(GetTime()-t)/nit);
#ifdef DEBUG
    cout << "y2 = " << y2 << endl;
#endif
    assert(y1==y2);

    t = GetTime();
    for (i = 0; i < nit; i++) TFast1Eval(f2, n-1, x, y2);
    t4 = (1000*(GetTime()-t)/nit);
#ifdef DEBUG
    cout << "f2 = " << f2 << endl;
#endif
    assert(f1==f2);

    t = GetTime();
    for (i = 0; i < nit; i++) Fast2Eval(y3, f, x);
    t5 = (1000*(GetTime()-t)/nit);
#ifdef DEBUG
    cout << "y3 = " << y3 << endl;
#endif
    assert(y1==y3);

    t = GetTime();
    for (i = 0; i < nit; i++) TFast2Eval(f3, n-1, x, y2);
    t6 = (1000*(GetTime()-t)/nit);
#ifdef DEBUG
    cout << "f3 = " << f3 << endl;
#endif
    assert(f1==f3);

    cout << n << "\t" << t1 << "\t" << t2 << "\t" << t2/t1 << "\t" << t3 << "\t" << t4 << "\t" << t4/t3 << "\t" << t5 << "\t" << t6 << "\t" << t6/t5 << endl;
    cout.flush();

  }
}

void test_interp(void) {

  double t, t1, t2, t3, t4;

  ZZ_pX f1,f2;
  vec_ZZ_p x,y,y1,y2;
  long i, k, n, nit;

#ifdef DEBUG
  long N = 0;
#else
  long N = 500000;
#endif

  cout << "# deg\tPlainInterp\tTPlainInterp\tTPlainInterp/PlainInterp\tFastInterp\ttFastInterp\tTFastInterp/FastInterp" << endl;

  for(n = 2; n < 200; n++) {
    if (n > 10) {
      n+=1;
      N = 50000;
    }

    x.SetLength(n);
    for(i = 0; i < n; i++) x[i]=i+1;
    y.SetLength(n);
    for(i = 0; i < n; i++) random(y[i]);

#ifdef DEBUG
    cout << "x = " << x << endl;
    cout << "y = " << y << endl;
#endif

    if (debug) nit = 1;
    else nit = 1 + N/n;

    t = GetTime();
    for (i = 0; i < nit; i++) PlainInterp(f1, x, y);
    t1 = (1000*(GetTime()-t)/nit);
#ifdef DEBUG
    cout << "f1 = " << f1 << endl;
#endif

    t = GetTime();
    for (i = 0; i < nit; i++) TPlainInterp(y1, x, f1);
    t2 = (1000*(GetTime()-t)/nit);
#ifdef DEBUG
    cout << "y1 = " << y1 << endl;
#endif

    t = GetTime();
    for (i = 0; i < nit; i++) FastInterp(f2, x, y);
    t3= (1000*(GetTime()-t)/nit);
#ifdef DEBUG
    cout << "f2 = " << f2 << endl;
#endif
    assert(f1==f2);

    t = GetTime();
    for (i = 0; i < nit; i++) TFastInterp(y2, x, f2);
    t4 = (1000*(GetTime()-t)/nit);
#ifdef DEBUG
    cout << "y2 = " << y2 << endl;
#endif
    assert(y1==y2);

    cout << n << "\t" << t1 << "\t" << t2 << "\t" << t2/t1 << "\t" << t3 << "\t" << t4 << "\t" << t4/t3 << endl;
    cout.flush();
  }
}

int main(int argc, char **argv) {

  if (argc != 4) {
    usage(argc,argv);
    exit(0);
  }

  debug = atol(argv[1]);
  cerr << "debug = " << debug << endl;

  long p_bit_size = atol(argv[2]);
  ZZ p; 
  RandomPrime(p,p_bit_size);
  cerr << "p = " << p << endl;
  ZZ_p::init(p);

  cout << "Generated by " << argv[0] << " " << argv[1] << " " << argv[2] << " " << argv[3] << endl;
  
  switch(atol(argv[3])) {
  case 0:
    test_mul();
    break;
  case 1:
    test_rem();
    break;
  case 2:
    test_eval();
    break;
  case 3:
    test_interp();
    break;
  default:
    usage(argc,argv);
    exit(0);
  }
}
